CON
{***************************************
 *  SPIO 8 x 8 led/keypad pasm driver  *
 *  Author: Kurt Winn March 2013       *
 *  Copyright (c) 2013 Kurt Winn       *
 *  See end of file for terms of use.  *
 ***************************************
}
bytecnt = 8

' **************************************************************************************************************************
' Constants required by shift register input/output subroutine 
' **************************************************************************************************************************

VAR
  
'**************************************************************************************************************************************
pub main(ParClkpin, SerClkpin, IOpin) | i
{
  Receive/send data from/to a p-s and a s-p chip for overlapped serial input and output.
  ParClkpin loads data to the parallel input ('165) and parallel output ('595, TPIC) registers.
  SerClkpin shifts the data in from the '165 and out to the '595 or TPIC.
  IOpin is the shared data input from the '165 and data output to the '595 or TPIC.
  A series current limiting resistor required between the '165 output and the prop IO pin.
  SwitchArray and LedArray are the addresses of the byte arrays where the input and output data are stored.
}
  Pclk := |< ParClkpin          ' set up paralell clock mask
  Sclk := |< SerClkpin          ' set up serial clock mask
  Datapin := IOpin              ' save pin number
  Datmask := |< IOpin           ' set up data input/output pin mask
  Inptr := @SwitchArray         ' set up pointer to hub input data table (switch data array)
  Outptr := @LedArray           ' set up pointer to hub output data table (led data array)

   cognew(@start, 0)

pub getIndata
    return(Indata)
pub getOutdata
    return(Outdata)
        
PUB setled(number, state) | index, mask                 ' set a led to on or off - leds are numbered 1 - 64 (row x col)
    index := ((number - 1) & $38) >> 3                        ' get the offset into the byte array
    mask := |< ((number - 1) & $7)
    if state == 1
      LedArray[index] |= mask                           ' set bit to 1
    else                                                                             
      LedArray[index] &= !mask                          ' set bit to 0
     
PUB setdig(digit, value)                                ' digits are numbered 0 - 16, 1 digit per row
    LedArray[digit - 1] := segtbl[value]                ' output 7 segment data to led digit

PUB setdp(digit)
    LedArray[digit - 1] |= %10000000                    ' set decimal point on

pub setrow(row, c)                                      ' set single byte of led array
    LedArray[row - 1] := c
        
pub getled(row)                                         ' returns a row of led data
    return LedArray[row - 1]

pub getsw(row)                                          ' get the status of a row of switches
    return SwitchArray[row - 1]

pub fill(c) | row                                       ' fill led array with byte
        repeat row from 0 to 7
          LedArray[row] := c
    
DAT
{ **********************************************************************************************************************
  Seven segment lookup table. Assumes column 1 to 7 corresponds to segment a to g respectively. Col. 8 is decimal point
  Arrays for storing led/digit data and input data from switches.      
  **********************************************************************************************************************}
segtbl  byte  %00111111 ' 0 $3F
        byte  %00000110 ' 1 $06
        byte  %01011011 ' 2 $5B
        byte  %01001111 ' 3 $4F
        byte  %01100110 ' 4 $66       
        byte  %01101101 ' 5 $6D
        byte  %01111101 ' 6 $7D
        byte  %00000111 ' 7 $07
        byte  %01111111 ' 8 $7F
        byte  %01100111 ' 9 $67
        byte  %01110111 ' A $77
        byte  %01111100 ' b $7C
        byte  %00111001 ' C $39
        byte  %01011110 ' d $5E
        byte  %01111001 ' E $79
        byte  %01110001 ' F $71
        byte  %00000000 ' blank $10

LedArray      byte 9,9,9,9,9,9,9,9
SwitchArray   byte 3,3,3,3,3,3,3,3       


{ **********************************************************************************************************************
  Initialization for multiplexing display. Sets up serial and parallel clock pins and clears display registers
  **********************************************************************************************************************}

        org   0
Start         or        outa, Pclk              ' set paralell clock pin high 
              or        outa, Sclk              ' set serial clock pin high 
              or        dira, Pclk              ' set paralell clock pin to output 
              or        dira, Sclk              ' set serial clock pin to output
              mov       Outdata, #$0
              call      #r_w8                   ' clear the shift registers
              call      #r_w8
              call      #load                   ' blank display

{ **********************************************************************************************************************
  Main loop for multiplexing display.
  - Set the number of bytes to output and initialize the row select bit
  - Output 8 bits of digit select data 
  - Read the byte of data to be displayed from hub and output it
  - Read the 8 bits of switch data for this row from the '165
  - Toggle the Pclk data to output the row data and row select bits for display
  - Adjust pointers and row select for next row or reset pointers if all rows done  
  **********************************************************************************************************************}
:loop1
'- Set the number of bytes to output and initialize the row select bit
              mov       temp1, #bytecnt         ' set number of bytes to output
              mov       temp2, #$01             ' set up row select bit for first row

:loop2
' - Output 8 bits of digit select data and input 8 bits from the current row of the switches
              mov       Outdata, temp2          ' put row select bit in Data
              call      #r_w8                   ' shift out the new row select bit

' - Read the byte of data to be displayed from hub, output it, and input 8 bits from the current row of switches
              rdbyte    Outdata, Outptr         ' read 8 bits to display on this row/digit
              call      #r_w8                   ' shift out the new column select (led) bits

'  - Toggle the Pclk data to output the row data and row select bits for display
              call      #load                   ' transfer row and column data to output register and switch data in to '165                   
              
'  - Write the 8 bits of switch data for this row from the '165 

              wrbyte    Indata, Inptr           ' write switch status for this row

         
'  - Adjust pointers and row select for next row or reset pointers if all rows done  
              add       Outptr, #1              ' increment output data pointer
              add       Inptr, #1               ' increment input data pointer
              shl       temp2, #1               ' shift row select bit for next row
              djnz      temp1, #:loop2          ' loop for next row if not done

              sub       Outptr, #bytecnt        ' reset pointer 
              sub       Inptr, #bytecnt         ' reset pointer
              jmp       #:loop1                 ' repeat from row 1

{ **********************************************************************************************************************
  Shift out 8 data bits in Outdata to '595/TPIC - MSB to LSB
  Shift in 8 data bits from the '165 at the same time
  **********************************************************************************************************************}
r_w8          ' writes 8 bits to '595/TPIC
              mov       temp3, #8               ' number of bits to output
              rev       Outdata, #24               ' bit order reversed
:loop1              
              ' input bit from '165
              test      datmask, ina wc         ' set/clear carry flag if input pin is high/low
              muxc      Indata, #1              ' input carry bit to bit 1 of Datin          ' 
              rol       Indata, #1              ' shift data for next bit

              ' output data bit to '595/TPIC 
              or        dira, Datmask           ' set input/output pin to output
              ror       Outdata, #1 wc          ' rotate bit and set carry if 1
              muxc      outa, Datmask           ' output the carry bit state to the pin
              call      #shift                  ' shift bit into serial register
              andn      dira, Datmask           ' set input/output pin to input                          
              djnz      temp3, #:loop1          ' repeat for 8 bits

r_w8_ret      ret

' ***********************************************************************************************************************
shift                   ' Toggle the serial clock to shift bits in the serial registers  
              call      #wait
              andn      outa, Sclk              ' set ParClk low - loads parallel data in to '165 serial shift register
              call      #wait
              or        outa, Sclk              ' set ParClk high - transfers data from '595/TPIC serial register to output register
              call      #wait
shift_ret     ret

' ***********************************************************************************************************************
load                    ' shift data in to P-S register, and out from the S-P serial register to the output register
              andn      outa, Pclk              ' set ParClk low - loads parallel data in to '165 serial shift register
              call      #wait
              or        outa, Pclk              ' set ParClk high - transfers data from '595/TPIC serial register to output register
              call      #wait
load_ret      ret

' ***********************************************************************************************************************
wait          mov       target, cnt
              add       target, delta
              waitcnt   target, delta
wait_ret      ret


Pclk    long  0         ' paralell clock pin mask
Sclk    long  0         ' serial clock pin mask
Datapin long  0         ' serial data input/output pin number
Datmask long  0         ' serial data input/output pin mask
Outptr  long  0         ' pointer to output data for '595/TPIC
Outdata long  0         ' data output to TPIC/'595
Inptr   long  0         ' pointer to input data from '165
Indata  long  0         ' data input from '165
temp1   long  0
temp2   long  0
temp3   long  0
temp4   long  0
target  long  0         
delta   long  2_000       
                            